from typing import Any, Callable, Dict, Union

from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path

import torch
from pytorch_lightning import Callback, Trainer

from lighter import LighterSystem


class LighterBaseWriter(ABC, Callback):
    """
    Base class for defining custom Writer. It provides the structure to save predictions in various formats.

    Subclasses should implement:
        1) `self.writers` attribute to specify the supported formats and their corresponding writer functions.
        2) `self.write()` method to specify the saving strategy for a prediction.

    Args:
        directory (str): Base directory for saving. A new sub-directory with current date and time will be created inside.
        writer (Union[str, Callable]): Name of the writer function registered in `self.writers`, or a custom writer function.
    """

    def __init__(self, directory: str, writer: Union[str, Callable]) -> None:
        """
        Initialize the LighterBaseWriter.

        Args:
            directory (str): Base directory for saving. A new sub-directory with current date and time will be created inside.
            writer (Union[str, Callable]): Name of the writer function registered in `self.writers`, or a custom writer function.
        """
        # Create a unique directory using the current date and time
        self.directory = Path(directory) / datetime.now().strftime("%Y%m%d_%H%M%S")

        # Check if the writer is a string and if it exists in the writers dictionary
        if isinstance(writer, str):
            if writer not in self.writers:
                raise ValueError(f"Writer for format {writer} does not exist. Available writers: {self.writers.keys()}.")
            self.writer = self.writers[writer]
        else:
            # If the writer is not a string, it is assumed to be a callable function
            self.writer = writer

        # Prediction counter. Used when IDs are not provided. Initialized in `self.setup()` based on the DDP rank.
        self._pred_counter = None

    @property
    @abstractmethod
    def writers(self) -> Dict[str, Callable]:
        """
        Property to define the default writer functions.
        """

    @abstractmethod
    def write(self, tensor: torch.Tensor, id: int) -> None:
        """
        Method to define how a tensor should be saved. The input tensor will be a single tensor without
        the batch dimension.

        For each supported format, there should be a corresponding writer function registered in `self.writers`
        A specific writer function can be retrieved using `self.get_writer(self.format)`.

        Args:
            tensor (torch.Tensor): tensor, without the batch dimension, to be saved.
            id (int): identifier for the tensor, can be used for naming files or adding table records.
        """

    def setup(self, trainer: Trainer, pl_module: LighterSystem, stage: str) -> None:
        """
        Callback function to set up necessary prerequisites: prediction count and prediction directory.
        When executing in a distributed environment, it ensures that:
        1. Each distributed node initializes a prediction count based on its rank.
        2. All distributed nodes write predictions to the same directory.
        3. The directory is accessible to all nodes, i.e., all nodes share the same storage.
        """
        if stage != "predict":
            return

        # Initialize the prediction count with the rank of the current process
        self._pred_counter = torch.distributed.get_rank() if trainer.world_size > 1 else 0

        # Ensure all distributed nodes write to the same directory
        self.directory = trainer.strategy.broadcast(self.directory, src=0)
        if trainer.is_global_zero:
            self.directory.mkdir(parents=True)
        # Wait for rank 0 to create the directory
        trainer.strategy.barrier()

        # Ensure all distributed nodes have access to the directory
        if not self.directory.exists():
            raise RuntimeError(
                f"Rank {trainer.global_rank} does not share storage with rank 0. Ensure nodes have common storage access."
            )

    def on_predict_batch_end(
        self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """
        Callback method executed at the end of each prediction batch/step.
        If the IDs are not provided, it generates global unique IDs based on the prediction count.
        Finally, it writes the predictions using the specified writer.
        """

        # If the IDs are not provided, generate global unique IDs based on the prediction count. DDP supported.
        if outputs["id"] is None:
            batch_size = len(outputs["pred"])
            world_size = trainer.world_size
            outputs["id"] = list(range(self._pred_counter, self._pred_counter + batch_size * world_size, world_size))
            self._pred_counter += batch_size * world_size

        for id, pred in zip(outputs["id"], outputs["pred"]):
            self.write(tensor=pred, id=id)



from typing import Any, Callable, Dict, List, Union

import inspect

from torch.optim.optimizer import Optimizer


def ensure_list(vals: Any) -> List:
    """Wrap the input into a list if it is not a list. If it is a None, return an empty list.

    Args:
        vals (Any): input to wrap into a list.

    Returns:
        List: output list.
    """
    if isinstance(vals, list):
        return vals
    if isinstance(vals, tuple):
        return list(vals)
    if vals is None:
        return []
    return [vals]


def ensure_dict_schema(input_dict: Dict[str, Any], schema: Dict[str, Any]) -> Dict[str, Any]:
    """
    Ensure that the input dict has the specified schema. If no value is set
    for a key in the input dict, the default value from the schema is used.
    This function supports nested dictionaries.

    Args:
        input_dict (Dict[str, Any]): The input dictionary to merge with the schema.
        schema (Dict[str, Any]): A schema dictionary with default values specified.

    Returns:
        Dict[str, Any]: The merged dictionary. If input_dict is None, returns the schema dictionary.

    Raises:
        ValueError: If the input dictionary has other keys than the specified schema keys.

    """
    output_dict = schema.copy()
    if input_dict is not None:
        for key, value in input_dict.items():
            if key not in schema:
                raise ValueError(f"Key {key} is not defined in the schema.")
            if isinstance(value, dict) and isinstance(schema[key], dict):
                output_dict[key] = ensure_dict_schema(value, schema[key])
            else:
                output_dict[key] = value
    return output_dict


def setattr_dot_notation(obj: Callable, attr: str, value: Any):
    """Set object's attribute. Supports dot notation.

    Args:
        obj (Callable): object.
        attr (str): attribute name of the object.
        value (Any): value to set the attribute to.
    """
    if "." not in attr:
        if not hasattr(obj, attr):
            raise AttributeError(f"`{get_name(obj, True)}` has no attribute `{attr}`.")
        setattr(obj, attr, value)
    # Solve recursively if the attribute is defined in dot-notation
    else:
        obj_name, attr = attr.split(".", maxsplit=1)
        setattr_dot_notation(getattr(obj, obj_name), attr, value)


def hasarg(_callable: Callable, arg_name: str) -> bool:
    """Check if a function, class, or method has an argument with the specified name.

    Args:
        _callable (Callable): function, class, or method to inspect.
        arg_name (str): argument name to check for.

    Returns:
        bool: `True` if the argument if the specified name exists.
    """

    args = inspect.signature(_callable).parameters.keys()
    return arg_name in args


def get_name(_callable: Callable, include_module_name: bool = False) -> str:
    """Get the name of an object, class or function.

    Args:
        _callable (Callable): object, class or function.
        include_module_name (bool, optional): whether to include the name of the module from
            which it comes. Defaults to False.

    Returns:
        str: name
    """
    name = type(_callable).__name__ if isinstance(_callable, object) else _callable.__name__
    if include_module_name:
        module = type(_callable).__module__ if isinstance(_callable, object) else _callable.__module__
        name = f"{module}.{name}"
    return name


def apply_fns(data: Any, fns: Union[Callable, List[Callable]]) -> Any:
    """Apply a function or a list of functions on the input.

    Args:
        data (Any): input to apply the function(s) on.
        fns (Union[Callable, List[Callable]]): function or list of functions to apply on the input.

    Returns:
        Any: output of the function(s).
    """
    for fn in ensure_list(fns):
        data = fn(data)
    return data


def get_optimizer_stats(optimizer: Optimizer) -> Dict[str, float]:
    """
    Extract learning rates and momentum values from each parameter group of the optimizer.

    Args:
        optimizer (Optimizer): A PyTorch optimizer.

    Returns:
        Dict[str, float]: Dictionary with formatted keys and values for learning rates and momentum.
    """
    stats_dict = {}
    for group_idx, group in enumerate(optimizer.param_groups):
        lr_key = f"optimizer/{optimizer.__class__.__name__}/lr"
        momentum_key = f"optimizer/{optimizer.__class__.__name__}/momentum"

        # Add group index to the key if there are multiple parameter groups
        if len(optimizer.param_groups) > 1:
            lr_key += f"/group{group_idx+1}"
            momentum_key += f"/group{group_idx+1}"

        # Extracting learning rate
        stats_dict[lr_key] = group["lr"]

        # Extracting momentum or betas[0] if available
        if "momentum" in group:
            stats_dict[momentum_key] = group["momentum"]
        if "betas" in group:
            stats_dict[momentum_key] = group["betas"][0]

    return stats_dict



from typing import Dict

import importlib
import sys
from dataclasses import dataclass, field
from pathlib import Path

from loguru import logger
from monai.utils.module import optional_import


@dataclass
class OptionalImports:
    """Dataclass for handling optional imports.

    This class provides a way to handle optional imports in a convenient manner.
    It allows importing modules that may or may not be available, and raises an ImportError if the module is not available.

    Example: ::
        from lighter.utils.dynamic_imports import OPTIONAL_IMPORTS
        writer = OPTIONAL_IMPORTS["tensorboard"].SummaryWriter()

    Attributes:
        imports (Dict[str, object]): A dictionary to store the imported modules.
    """

    imports: Dict[str, object] = field(default_factory=dict)

    def __getitem__(self, module_name: str):
        """Get the imported module by name.

        Args:
            module_name (str): The name of the module to import.

        Raises:
            ImportError: If the module is not available.

        Returns:
            object: The imported module.
        """
        if module_name not in self.imports:
            self.imports[module_name], module_available = optional_import(module_name)
            if not module_available:
                raise ImportError(f"'{module_name}' is not available. Make sure that it is installed and spelled correctly.")
        return self.imports[module_name]


OPTIONAL_IMPORTS = OptionalImports()


def import_module_from_path(module_name: str, module_path: str) -> None:
    """Import a module from a given path and assign it a specified name.

    This function imports a module from the specified path and assigns it the specified name.

    Args:
        module_name (str): The name to assign to the imported module.
        module_path (str): The path to the module to import.

    Raises:
        ValueError: If the module has already been imported.
        FileNotFoundError: If the `__init__.py` file is not found in the module path.
    """
    # Based on https://stackoverflow.com/a/41595552.

    if module_name in sys.modules:
        raise ValueError(f"{module_name} has already been imported as module.")

    module_path = Path(module_path).resolve() / "__init__.py"
    if not module_path.is_file():
        raise FileNotFoundError(f"No `__init__.py` in `{module_path}`.")
    spec = importlib.util.spec_from_file_location(module_name, str(module_path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[module_name] = module
    logger.info(f"{module_path.parent} imported as '{module_name}' module.")



from typing import Callable, Dict, Union

from functools import partial
from pathlib import Path

import torch
import torchvision
from monai.data import metatensor_to_itk_image
from monai.transforms import DivisiblePad

from lighter.callbacks.utils import preprocess_image
from lighter.callbacks.writer.base import LighterBaseWriter
from lighter.utils.dynamic_imports import OPTIONAL_IMPORTS


class LighterFileWriter(LighterBaseWriter):
    """
    Writer for writing predictions to files. Supports writing tensors, images, videos, and ITK images.
    To write to other formats, a custom writer function can be provided to the `writer` argument or,
    for a more permanent solution, it can be added to the `self.writers` dictionary.

    Args:
        directory (Union[str, Path]): The directory where the files should be written.
        writer (Union[str, Callable]): Name of the writer function registered in `self.writers` or a custom writer function.
            Available writers: "tensor", "image", "video", "itk_nrrd", "itk_seg_nrrd", "itk_nifti".
            A custom writer function must take two arguments: `path` and `tensor`, and write the tensor to the specified path.
            `tensor` is a single tensor without the batch dimension.
    """

    def __init__(self, directory: Union[str, Path], writer: Union[str, Callable]) -> None:
        super().__init__(directory, writer)

    @property
    def writers(self) -> Dict[str, Callable]:
        return {
            "tensor": write_tensor,
            "image": write_image,
            "video": write_video,
            "itk_nrrd": partial(write_itk_image, suffix=".nrrd"),
            "itk_seg_nrrd": partial(write_itk_image, suffix=".seg.nrrd"),
            "itk_nifti": partial(write_itk_image, suffix=".nii.gz"),
        }

    def write(self, tensor: torch.Tensor, id: Union[int, str]) -> None:
        """
        Write the tensor to the specified path in the given format.

        Args:
            tensor (Tensor): tensor, without the batch dimension, to be written.
            id (Union[int, str]): identifier, used for file-naming.
            format (str): format in which tensor should be written.
        """
        # Determine the path for the file based on prediction count. The suffix must be added by the writer function.
        path = self.directory / str(id)
        path.parent.mkdir(exist_ok=True, parents=True)
        # Write the tensor to the file.
        self.writer(path, tensor)


def write_tensor(path, tensor):
    torch.save(tensor, path.with_suffix(".pt"))


def write_image(path, tensor):
    path = path.with_suffix(".png")
    tensor = preprocess_image(tensor)
    torchvision.io.write_png(tensor, path)


def write_video(path, tensor):
    path = path.with_suffix(".mp4")
    # Video tensor must be divisible by 2. Pad the height and width.
    tensor = DivisiblePad(k=(0, 2, 2), mode="minimum")(tensor)
    # Video tensor must be THWC. Permute CTHW -> THWC.
    tensor = tensor.permute(1, 2, 3, 0)
    # Video tensor must have 3 channels (RGB). Repeat the channel dim to convert grayscale to RGB.
    if tensor.shape[-1] == 1:
        tensor = tensor.repeat(1, 1, 1, 3)
    # Video tensor must be in the range [0, 1]. Scale to [0, 255].
    tensor = (tensor * 255).to(torch.uint8)
    torchvision.io.write_video(str(path), tensor, fps=24)


def write_itk_image(path: str, tensor: torch.Tensor, suffix) -> None:
    path = path.with_suffix(suffix)
    itk_image = metatensor_to_itk_image(tensor, channel_dim=0, dtype=tensor.dtype)
    OPTIONAL_IMPORTS["itk"].imwrite(itk_image, str(path), True)



from functools import partial

import fire

from lighter.utils.runner import run_trainer_method


def interface():
    """Defines the command line interface for running Trainer's methods. The available methods are:
    - fit
    - validate
    - test
    - predict
    - tune
    """
    commands = {}
    # All Trainer methods' names
    for name in ["fit", "validate", "test", "predict", "tune"]:
        # Creates command for fire to run. The command is the Trainer method's name. Fire passes
        # **kwargs to the function.
        commands[name] = partial(run_trainer_method, name)

    fire.Fire(commands)



from typing import Any, Callable, Dict, Union

import itertools
from pathlib import Path

import pandas as pd
import torch
from pytorch_lightning import Trainer

from lighter import LighterSystem
from lighter.callbacks.writer.base import LighterBaseWriter


class LighterTableWriter(LighterBaseWriter):
    """
    Writer for saving predictions in a table format.

    Args:
        directory (Path): The directory where the CSV will be saved.
        writer (Union[str, Callable]): Name of the writer function registered in `self.writers` or a custom writer function.
            Available writers: "tensor". A custom writer function must take a single argument: `tensor`, and return the record
            to be saved in the CSV file. The tensor will be a single tensor without the batch dimension.
    """

    def __init__(self, directory: Union[str, Path], writer: Union[str, Callable]) -> None:
        super().__init__(directory, writer)
        self.csv_records = {}

    @property
    def writers(self) -> Dict[str, Callable]:
        return {
            "tensor": lambda tensor: tensor.tolist(),
        }

    def write(self, tensor: Any, id: Union[int, str]) -> None:
        """
        Write the tensor as a table record using the specified writer.

        Args:
            tensor (Any): tensor, without the batch dimension, to be recorded.
            id (Union[int, str]): identifier, used as the key for the record.
        """
        column = "pred"
        record = self.writer(tensor)
        self.csv_records.setdefault(id, {})[column] = record

    def on_predict_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        """
        Callback invoked at the end of the prediction epoch to save predictions to a CSV file.

        This method is responsible for organizing prediction records and saving them as a CSV file.
        If training was done in a distributed setting, it gathers predictions from all processes
        and then saves them from the rank 0 process.
        """
        csv_path = self.directory / "predictions.csv"

        # Sort the records by ID and convert the dictionary to a list
        self.csv_records = [self.csv_records[id] for id in sorted(self.csv_records)]

        # If in distributed data parallel mode, gather records from all processes to rank 0.
        if trainer.world_size > 1:
            # Create a list to hold the records from each process. Used on rank 0 only.
            gather_csv_records = [None] * trainer.world_size if trainer.is_global_zero else None
            # Each process sends its records to rank 0, which stores them in the `gather_csv_records`.
            torch.distributed.gather_object(self.csv_records, gather_csv_records, dst=0)
            # Concatenate the gathered records
            if trainer.is_global_zero:
                self.csv_records = list(itertools.chain(*gather_csv_records))

        # Save the records to a CSV file
        if trainer.is_global_zero:
            pd.DataFrame(self.csv_records).to_csv(csv_path)



from typing import Any, Dict

from monai.bundle.config_parser import ConfigParser
from pytorch_lightning import seed_everything

from lighter.utils.dynamic_imports import import_module_from_path


def parse_config(**kwargs):
    """
    Parses configuration files and updates the provided parser
    with given keyword arguments. Returns an updated parser object.

    Args:
        **kwargs (dict): Keyword arguments containing configuration data.
            config_file (str): Path to the main configuration file.
            args_file (str, optional): Path to secondary configuration file for additional arguments.
            Additional key-value pairs can also be provided to be added or updated in the parser.

    Returns:
        ConfigParser: An instance of ConfigParser with parsed and merged configuration data.
    """

    # Check that a config file is specified.
    if "config_file" not in kwargs:
        raise ValueError("--config_file not specified. Exiting.")

    # Parse the config file(s).
    parser = ConfigParser()
    parser.read_config(kwargs.pop("config_file"))
    parser.update(pairs=kwargs)

    # Import the project folder as a module, if specified.
    project = parser.get("project", None)
    if project is not None:
        import_module_from_path("project", project)

    return parser


def run_trainer_method(method: Dict, **kwargs: Any):
    """Call monai.bundle.run() on a Trainer method. If a project path
    is defined in the config file(s), import it.

    Args:
        method_name: name of the Trainer method to run. ["fit", "validate", "test", "predict", "tune"].
        **kwargs (Any): keyword arguments passed to the `monai.bundle.run` function.
    """
    # Sets the random seed to `PL_GLOBAL_SEED` env variable. If not specified, it picks a random seed.
    seed_everything()

    # Parse the config file(s).
    parser = parse_config(**kwargs)

    # Get trainer and system
    trainer = parser.get_parsed_content("trainer")
    system = parser.get_parsed_content("system")

    config = parser.get()
    config.pop("_meta_")
    # Save the config to model checkpoints under the "hyper_parameters" key.
    system.save_hyperparameters(config)
    # Log the config.
    if trainer.logger is not None:
        trainer.logger.log_hyperparams(config)

    # Run the Trainer method.
    if not hasattr(trainer, method):
        raise ValueError(f"Trainer has no method named {method}.")
    getattr(trainer, method)(system)



from typing import Any, List, Optional, Union

from loguru import logger
from pytorch_lightning import Callback, Trainer
from torch.nn import Module

from lighter import LighterSystem
from lighter.utils.misc import ensure_list


class LighterFreezer(Callback):
    """
    Callback to freeze the parameters/layers of a model. Can be run indefinitely or until a specified step or epoch.
    `names` and`name_starts_with` can be used to specify which parameters to freeze.
    If both are specified, the parameters that match any of the two will be frozen.

    Args:
        names (str, List[str], optional): The names of the parameters to be frozen. Defaults to None.
        name_starts_with (str, List[str], optional): The prefixes of the parameter names to be frozen. Defaults to None.
        except_names (str, List[str], optional): The names of the parameters to be excluded from freezing. Defaults to None.
        except_name_starts_with (str, List[str], optional): The prefixes of the parameter names to be excluded from freezing.
            Defaults to None.
        until_step (int, optional): The maximum step to freeze parameters until. Defaults to None.
        until_epoch (int, optional): The maximum epoch to freeze parameters until. Defaults to None.

    Raises:
        ValueError: If neither `names` nor `name_starts_with` are specified.
        ValueError: If both `until_step` and `until_epoch` are specified.

    """

    def __init__(
        self,
        names: Optional[Union[str, List[str]]] = None,
        name_starts_with: Optional[Union[str, List[str]]] = None,
        except_names: Optional[Union[str, List[str]]] = None,
        except_name_starts_with: Optional[Union[str, List[str]]] = None,
        until_step: int = None,
        until_epoch: int = None,
    ) -> None:
        super().__init__()

        if names is None and name_starts_with is None:
            raise ValueError("At least one of `names` or `name_starts_with` must be specified.")

        if until_step is not None and until_epoch is not None:
            raise ValueError("Only one of `until_step` or `until_epoch` can be specified.")

        self.names = ensure_list(names)
        self.name_starts_with = ensure_list(name_starts_with)
        self.except_names = ensure_list(except_names)
        self.except_name_starts_with = ensure_list(except_name_starts_with)
        self.until_step = until_step
        self.until_epoch = until_epoch

        self._frozen_state = False

    def on_train_batch_start(self, trainer: Trainer, pl_module: LighterSystem, batch: Any, batch_idx: int) -> None:
        self._on_batch_start(trainer, pl_module)

    def on_validation_batch_start(
        self, trainer: Trainer, pl_module: LighterSystem, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        self._on_batch_start(trainer, pl_module)

    def on_test_batch_start(
        self, trainer: Trainer, pl_module: LighterSystem, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        self._on_batch_start(trainer, pl_module)

    def on_predict_batch_start(
        self, trainer: Trainer, pl_module: LighterSystem, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        self._on_batch_start(trainer, pl_module)

    def _on_batch_start(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        """
        Freezes the parameters of the model at the start of each training batch.

        Args:
            trainer (Trainer): The PyTorch Lightning trainer instance.
            pl_module (LighterSystem): The PyTorch Lightning module instance.
            batch (Any): The current batch data.
            batch_idx (int): The index of the current batch.
            dataloader_idx (int): The index of the current dataloader.

        """
        current_step = trainer.global_step
        current_epoch = trainer.current_epoch

        if self.until_step is not None and current_step >= self.until_step:
            if self._frozen_state:
                logger.info(f"Reached step {self.until_step} - unfreezing the previously frozen layers.")
                self._set_model_requires_grad(pl_module, True)
            return

        if self.until_epoch is not None and current_epoch >= self.until_epoch:
            if self._frozen_state:
                logger.info(f"Reached epoch {self.until_epoch} - unfreezing the previously frozen layers.")
                self._set_model_requires_grad(pl_module, True)
            return

        if not self._frozen_state:
            self._set_model_requires_grad(pl_module, False)

    def _set_model_requires_grad(self, model: Union[Module, LighterSystem], requires_grad: bool) -> None:
        """
        Sets the requires_grad attribute of the model's parameters.

        Args:
            model (Module): The PyTorch model whose parameters need to be frozen.
            requires_grad (bool): Whether to freeze the parameters or not.

        """
        # If the model is a `LighterSystem`, get the underlying PyTorch model.
        if isinstance(model, LighterSystem):
            model = model.model

        frozen_layers = []
        # Freeze the specified parameters.
        for name, param in model.named_parameters():
            # Leave the excluded-from-freezing parameters trainable.
            if self.except_names and name in self.except_names:
                param.requires_grad = True
                continue
            if self.except_name_starts_with and any(name.startswith(prefix) for prefix in self.except_name_starts_with):
                param.requires_grad = True
                continue

            # Freeze/unfreeze the specified parameters, based on the `requires_grad` argument.
            if self.names and name in self.names:
                param.requires_grad = requires_grad
                frozen_layers.append(name)
                continue
            if self.name_starts_with and any(name.startswith(prefix) for prefix in self.name_starts_with):
                param.requires_grad = requires_grad
                frozen_layers.append(name)
                continue

            # Otherwise, leave the parameter trainable.
            param.requires_grad = True

        self._frozen_state = not requires_grad
        # Log only when freezing the parameters.
        if self._frozen_state:
            logger.info(
                f"Setting requires_grad={requires_grad} the following layers"
                + (f" until step {self.until_step}" if self.until_step is not None else "")
                + (f" until epoch {self.until_epoch}" if self.until_epoch is not None else "")
                + f": {frozen_layers}"
            )



import torch
import torchvision


def get_lighter_mode(lightning_stage: str) -> str:
    """Converts the name of a PyTorch Lightnig stage to the name of its corresponding Lighter mode.

    Args:
        lightning_stage (str): stage in which PyTorch Lightning Trainer is. Can be accessed using `trainer.state.stage`.

    Returns:
        str: name of the Lighter mode.
    """
    lightning_to_lighter = {"train": "train", "validate": "val", "test": "test"}
    return lightning_to_lighter[lightning_stage]


def preprocess_image(image: torch.Tensor) -> torch.Tensor:
    """Preprocess the image before logging it. If it is a batch of multiple images,
    it will create a grid image of them. In case of 3D, a single image is displayed
    with slices stacked vertically, while a batch of 3D images as a grid where each
    column is a different 3D image.
    Args:
        image (torch.Tensor): 2D or 3D image tensor.
    Returns:
        torch.Tensor: image ready for logging.
    """
    # If 3D (BCDHW), concat the images vertically and horizontally.
    if image.ndim == 5:
        shape = image.shape
        # BCDHW -> BC(D*H)W. Combine slices of a 3D images vertically into a single 2D image.
        image = image.view(shape[0], shape[1], shape[2] * shape[3], shape[4])
        # BCDHW -> 1CDH(B*W). Concat images in the batch horizontally, and unsqueeze to add back the B dim.
        image = torch.cat([*image], dim=-1).unsqueeze(0)
    # If only one image in the batch, select it and return it. Same happens when the images are 3D as they
    # are combined into a single image. `make_grid` is called when a batch of multiple 2D image is provided.
    return image[0] if image.shape[0] == 1 else torchvision.utils.make_grid(image, nrow=8)



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from functools import partial

import pytorch_lightning as pl
import torch
from loguru import logger
from torch.nn import Module, ModuleDict
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader, Dataset, Sampler
from torchmetrics import Metric, MetricCollection

from lighter.utils.collate import collate_replace_corrupted
from lighter.utils.misc import apply_fns, ensure_dict_schema, get_name, get_optimizer_stats, hasarg


class LighterSystem(pl.LightningModule):
    """_summary_

    Args:
        model (Module): The model.
        batch_size (int): Batch size.
        drop_last_batch (bool, optional): Whether the last batch in the dataloader should be dropped. Defaults to False.
        num_workers (int, optional): Number of dataloader workers. Defaults to 0.
        pin_memory (bool, optional): Whether to pin the dataloaders memory. Defaults to True.
        optimizer (Optimizer, optional): Optimizers. Defaults to None.
        scheduler (LRScheduler, optional): Learning rate scheduler. Defaults to None.
        criterion (Callable, optional): Criterion/loss function. Defaults to None.
        datasets (Dict[str, Dataset], optional): Datasets for train, val, test, and predict. Defaults to None.
        samplers (Dict[str, Sampler], optional): Samplers for train, val, test, and predict. Defaults to None.
        collate_fns (Dict[str, Union[Callable, List[Callable]]], optional):
            Collate functions for train, val, test, and predict. Defaults to None.
        metrics (Dict[str, Union[Metric, List[Metric], Dict[str, Metric]]], optional):
            Metrics for train, val, and test. Supports a single metric or a list/dict of `torchmetrics` metrics.
            Defaults to None.
        postprocessing (Dict[str, Union[Callable, List[Callable]]], optional):
            Postprocessing functions for input, target, and pred, for three stages - criterion, metrics,
            and logging. The postprocessing is done before each stage - for example, criterion postprocessing
            will be done prior to loss calculation. Note that the postprocessing of a latter stage stacks on
            top of the previous one(s) - for example, the logging postprocessing will be done on the data that
            has been postprocessed for the criterion and metrics earlier. Defaults to None.
        inferer (Callable, optional): The inferer must be a class with a `__call__` method that accepts two
            arguments - the input to infer over, and the model itself. Used in 'val', 'test', and 'predict'
            mode, but not in 'train'. Typically, an inferer is a sliding window or a patch-based inferer
            that will infer over the smaller parts of the input, combine them, and return a single output.
            The inferers provided by MONAI cover most of such cases (https://docs.monai.io/en/stable/inferers.html).
            Defaults to None.
    """

    def __init__(
        self,
        model: Module,
        batch_size: int,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        pin_memory: bool = True,
        optimizer: Optional[Optimizer] = None,
        scheduler: Optional[LRScheduler] = None,
        criterion: Optional[Callable] = None,
        datasets: Dict[str, Dataset] = None,
        samplers: Dict[str, Sampler] = None,
        collate_fns: Dict[str, Union[Callable, List[Callable]]] = None,
        metrics: Dict[str, Union[Metric, List[Metric], Dict[str, Metric]]] = None,
        postprocessing: Dict[str, Union[Callable, List[Callable]]] = None,
        inferer: Optional[Callable] = None,
    ) -> None:
        super().__init__()
        # Bypass LightningModule's check for default methods. We define them in self.setup().
        self._init_placeholders_for_dataloader_and_step_methods()

        # Model setup
        self.model = model
        self.batch_size = batch_size
        self.drop_last_batch = drop_last_batch

        # Criterion, optimizer, and scheduler
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler

        # DataLoader specifics
        self.num_workers = num_workers
        self.pin_memory = pin_memory

        # Datasets, samplers, and collate functions
        self.datasets = self._init_datasets(datasets)
        self.samplers = self._init_samplers(samplers)
        self.collate_fns = self._init_collate_fns(collate_fns)

        # Metrics
        self.metrics = self._init_metrics(metrics)

        # Postprocessing
        self.postprocessing = self._init_postprocessing(postprocessing)

        # Inferer for val, test, and predict
        self.inferer = inferer

        # Checks
        self._lightning_module_methods_defined = False
        self._target_not_used_reported = False
        self._batch_type_reported = False

    def forward(self, input: Union[torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor], Dict[str, torch.Tensor]]) -> Any:
        """Forward pass. Multi-input models are supported.

        Args:
            input (torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor], Dict[str, torch.Tensor]): input to the model.

        Returns:
            Any: output of the model.
        """

        # Keyword arguments to pass to the forward method
        kwargs = {}
        # Add `epoch` argument if forward accepts it
        if hasarg(self.model.forward, "epoch"):
            kwargs["epoch"] = self.current_epoch
        # Add `step` argument if forward accepts it
        if hasarg(self.model.forward, "step"):
            kwargs["step"] = self.global_step

        return self.model(input, **kwargs)

    def _base_step(self, batch: Union[List, Tuple], batch_idx: int, mode: str) -> Union[Dict[str, Any], Any]:
        """Base step for all modes ("train", "val", "test", "predict")

        Args:
            batch (List, Tuple):
                output of the DataLoader and input to the model.
            batch_idx (int): index of the batch. Not used, but PyTorch Lightning requires it.
            mode (str): mode in which the system is.

        Returns:
            Union[Dict[str, Any], Any]: For the training, validation and test step, it returns
                a dict containing loss, metrics, input, target, and pred. Loss will be `None`
                for the test step. Metrics will be `None` if no metrics are specified.

                For predict step, it returns pred only.
        """
        # Batch type check:
        # - Dict: must contain "input" and "target" keys, and optionally "id" key.
        if isinstance(batch, dict):
            if set(batch.keys()) not in [{"input", "target"}, {"input", "target", "id"}]:
                raise ValueError(
                    "A batch dict must have 'input', 'target', and, "
                    f"optionally 'id', as keys, but found {list(batch.keys())}"
                )
            batch["id"] = None if "id" not in batch else batch["id"]
        # - List/tuple: must contain two elements - input and target. After the check, convert it to dict.
        elif isinstance(batch, (list, tuple)):
            if len(batch) != 2:
                raise ValueError(
                    f"A batch must consist of 2 elements - input and target. However, {len(batch)} "
                    "elements were found. Note: if target does not exist, return `None` as target."
                )
            batch = {"input": batch[0], "target": batch[1], "id": None}
        # - Other types are not allowed.
        else:
            raise TypeError(
                "A batch must be a list, a tuple, or a dict."
                "A batch dict must have 'input' and 'target' keys, and optionally 'id'."
                "A batch list or a tuple must have 2 elements - input and target."
                "If target does not exist, return `None` as target."
            )

        # Split the batch into input, target, and id.
        input, target, id = batch["input"], batch["target"], batch["id"]

        # Forward
        if self.inferer and mode in ["val", "test", "predict"]:
            pred = self.inferer(input, self)
        else:
            pred = self(input)

        # Postprocessing for loss calculation.
        input = apply_fns(input, self.postprocessing["criterion"]["input"])
        target = apply_fns(target, self.postprocessing["criterion"]["target"])
        pred = apply_fns(pred, self.postprocessing["criterion"]["pred"])

        # Predict mode stops here.
        if mode == "predict":
            # Postprocessing for logging/writing.
            pred = apply_fns(pred, self.postprocessing["logging"]["pred"])
            return {"pred": pred, "id": id}

        # Calculate the loss.
        loss = self._calculate_loss(pred, target) if mode in ["train", "val"] else None

        # Postprocessing for metrics.
        input = apply_fns(input, self.postprocessing["metrics"]["input"])
        target = apply_fns(target, self.postprocessing["metrics"]["target"])
        pred = apply_fns(pred, self.postprocessing["metrics"]["pred"])

        # Calculate the step metrics. # TODO: Remove the "_" prefix when fixed https://github.com/pytorch/pytorch/issues/71203
        metrics = self.metrics["_" + mode](pred, target) if self.metrics["_" + mode] is not None else None

        # Postprocessing for logging/writing.
        input = apply_fns(input, self.postprocessing["logging"]["input"])
        target = apply_fns(target, self.postprocessing["logging"]["target"])
        pred = apply_fns(pred, self.postprocessing["logging"]["pred"])

        # Logging
        self._log_loss_metrics_and_optimizer(loss, metrics, self.optimizer, mode, batch_idx)

        return {"loss": loss, "metrics": metrics, "input": input, "target": target, "pred": pred, "id": id}

    def _calculate_loss(
        self, pred: Union[torch.Tensor, List, Tuple, Dict], target: Union[torch.Tensor, List, Tuple, Dict, None]
    ) -> torch.Tensor:
        """Calculates the loss.
        The method handles cases where the criterion function does not accept a `target` argument. If the criterion
        function does not accept a `target` argument, the LighterSystem passes only the predicted values to the criterion.

        Args:
            pred (torch.Tensor, List, Tuple, Dict): the predicted values from the model.
            target (torch.Tensor, List, Tuple, Dict, None): the target/label.

        Returns:
            torch.Tensor: the calculated loss.
        """

        # Keyword arguments to pass to the loss/criterion function
        kwargs = {}
        # Add `target` argument if forward accepts it.
        if hasarg(self.criterion.forward, "target"):
            kwargs["target"] = target
        else:
            if not self._target_not_used_reported and not self.trainer.sanity_checking:
                self._target_not_used_reported = True
                logger.info(
                    f"The criterion `{get_name(self.criterion, True)}` "
                    "has no `target` argument. In such cases, the LighterSystem "
                    "passes only the predicted values to the criterion. "
                    "If this is not the behavior you expected, redefine your "
                    "criterion so that it has a `target` argument."
                )
        return self.criterion(pred, **kwargs)

    def _log_loss_metrics_and_optimizer(
        self, loss: torch.Tensor, metrics: MetricCollection, optimizer: Optimizer, mode: str, batch_idx: int
    ) -> None:
        """
        Logs the loss, metrics, and optimizer statistics.

        Args:
            loss (torch.Tensor): The calculated loss.
            metrics (MetricCollection): The calculated metrics.
            optimizer (Optimizer): The optimizer used in the model.
            mode (str): The mode of operation (train/val/test/predict).
            batch_idx (int): The index of the current batch.
        """
        if self.trainer.logger is None:
            return

        default_kwargs = {"logger": True, "batch_size": self.batch_size}
        step_kwargs = {"on_epoch": False, "on_step": True}
        epoch_kwargs = {"on_epoch": True, "on_step": False}
        # Loss
        if loss is not None:
            self.log(f"{mode}/loss/step", loss, **default_kwargs, **step_kwargs)
            self.log(f"{mode}/loss/epoch", loss, **default_kwargs, **epoch_kwargs, sync_dist=True)
        # Metrics
        if metrics is not None:
            for k, v in metrics.items():
                self.log(f"{mode}/metrics/{k}/step", v, **default_kwargs, **step_kwargs)
                self.log(f"{mode}/metrics/{k}/epoch", v, **default_kwargs, **epoch_kwargs, sync_dist=True)
        # Optimizer's lr, momentum, beta. Logged in train mode and once per epoch.
        if mode == "train" and batch_idx == 0:
            for k, v in get_optimizer_stats(self.optimizer).items():
                self.log(f"{mode}/{k}", v, **default_kwargs, **epoch_kwargs)

    def _base_dataloader(self, mode: str) -> DataLoader:
        """Instantiate the dataloader for a mode (train/val/test/predict).
        Includes a collate function that enables the DataLoader to replace
        None's (alias for corrupted examples) in the batch with valid examples.
        To make use of it, write a try-except in your Dataset that handles
        corrupted data by returning None instead.

        Args:
            mode (str): mode for which to create the dataloader ["train", "val", "test", "predict"].

        Returns:
            DataLoader: instantiated DataLoader.
        """
        dataset = self.datasets[mode]
        sampler = self.samplers[mode]
        collate_fn = self.collate_fns[mode]

        if dataset is None:
            raise ValueError(f"Please specify '{mode}' dataset in the 'datasets' key of the config.")

        # Batch size is 1 when using an inference for two reasons:
        # 1) Inferer separates an input into multiple parts, forming a batch of its own.
        # 2) The val/test/pred data usually differ in their shape, so they cannot be stacked into a batch.
        batch_size = self.batch_size
        if self.inferer is not None and mode in ["val", "test", "predict"]:
            logger.info(f"Setting the '{mode}' mode dataloader to batch size of 1 because an inferer is provided.")
            batch_size = 1

        # A dataset can return None when a corrupted example occurs. This collate
        # function replaces None's with valid examples from the dataset.
        collate_fn = partial(collate_replace_corrupted, dataset=dataset, default_collate_fn=collate_fn)
        return DataLoader(
            dataset,
            sampler=sampler,
            shuffle=(mode == "train" and sampler is None),
            batch_size=batch_size,
            drop_last=(self.drop_last_batch if mode == "train" else False),
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            collate_fn=collate_fn,
        )

    def configure_optimizers(self) -> Dict:
        """LightningModule method. Returns optimizers and, if defined, schedulers.

        Returns:
            Dict: optimizer and, if defined, scheduler.
        """
        if self.optimizer is None:
            raise ValueError("Please specify 'system.optimizer' in the config.")
        if self.scheduler is None:
            return {"optimizer": self.optimizer}
        else:
            return {"optimizer": self.optimizer, "lr_scheduler": self.scheduler}

    def setup(self, stage: str) -> None:
        """Automatically called by the LightningModule after the initialization.
        `LighterSystem`'s setup checks if the required dataset is provided in the config and
        sets up LightningModule methods for the stage in which the system is.

        Args:
            stage (str): passed by PyTorch Lightning. ["fit", "validate", "test"].
        """
        # Stage-specific PyTorch Lightning methods. Defined dynamically so that the system
        # only has methods used in the stage and for which the configuration was provided.
        if not self._lightning_module_methods_defined:
            del (
                self.train_dataloader,
                self.training_step,
                self.val_dataloader,
                self.validation_step,
                self.test_dataloader,
                self.test_step,
                self.predict_dataloader,
                self.predict_step,
            )
            # Prevents the methods from being defined again. This is needed because `Trainer.tune()`
            # calls the `self.setup()` method whenever it runs for a new parameter.
            self._lightning_module_methods_defined = True

        # Training methods.
        if stage in ["fit", "tune"]:
            self.train_dataloader = partial(self._base_dataloader, mode="train")
            self.training_step = partial(self._base_step, mode="train")

        # Validation methods. Required in 'validate' stage and optionally in 'fit' or 'tune' stage.
        if stage == "validate" or (stage in ["fit", "tune"] and self.datasets["val"] is not None):
            self.val_dataloader = partial(self._base_dataloader, mode="val")
            self.validation_step = partial(self._base_step, mode="val")

        # Test methods.
        if stage == "test":
            self.test_dataloader = partial(self._base_dataloader, mode="test")
            self.test_step = partial(self._base_step, mode="test")

        # Predict methods.
        if stage == "predict":
            self.predict_dataloader = partial(self._base_dataloader, mode="predict")
            self.predict_step = partial(self._base_step, mode="predict")

    def _init_placeholders_for_dataloader_and_step_methods(self) -> None:
        """
        Initializes placeholders for dataloader and step methods.

        `LighterSystem` dynamically defines the `..._dataloader()`and `..._step()` methods
        in the `self.setup()` method. However, when `LightningModule` excepts them to be defined
        at init. To prevent it from throwing an error, the `..._dataloader()` and `..._step()`
        are initially defined as `lambda: None`, before `self.setup()` is called.
        """
        self.train_dataloader = self.training_step = lambda: None
        self.val_dataloader = self.validation_step = lambda: None
        self.test_dataloader = self.test_step = lambda: None
        self.predict_dataloader = self.predict_step = lambda: None

    def _init_datasets(self, datasets: Dict[str, Optional[Dataset]]):
        """Ensures that the datasets have the predefined schema."""
        return ensure_dict_schema(datasets, {"train": None, "val": None, "test": None, "predict": None})

    def _init_samplers(self, samplers: Dict[str, Optional[Sampler]]):
        """Ensures that the samplers have the predefined schema"""
        return ensure_dict_schema(samplers, {"train": None, "val": None, "test": None, "predict": None})

    def _init_collate_fns(self, collate_fns: Dict[str, Optional[Callable]]):
        """Ensures that the collate functions have the predefined schema."""
        return ensure_dict_schema(collate_fns, {"train": None, "val": None, "test": None, "predict": None})

    def _init_metrics(self, metrics: Dict[str, Optional[Union[Metric, List[Metric], Dict[str, Metric]]]]):
        """Ensures that the metrics have the predefined schema. Wraps each mode's metrics in
        a MetricCollection, and finally registers them with PyTorch using a ModuleDict.
        """
        metrics = ensure_dict_schema(metrics, {"train": None, "val": None, "test": None})
        for mode, metric in metrics.items():
            metrics[mode] = MetricCollection(metric) if metric is not None else None
        # TODO: Remove the prefix addition line below when fixed https://github.com/pytorch/pytorch/issues/71203
        metrics = {f"_{k}": v for k, v in metrics.items()}
        return ModuleDict(metrics)

    def _init_postprocessing(self, postprocessing: Dict[str, Optional[Union[Callable, List[Callable]]]]):
        """Ensures that the postprocessing functions have the predefined schema."""
        subschema = {"input": None, "target": None, "pred": None}
        schema = {"criterion": subschema, "metrics": subschema, "logging": subschema}
        return ensure_dict_schema(postprocessing, schema)



from typing import Dict, List

import torch
from loguru import logger
from torch.nn import Identity, Module, Sequential

from lighter.utils.misc import setattr_dot_notation


def replace_layer_with(model: Module, layer_name: str, new_layer: Module) -> Module:
    """Replaces the specified layer of the network with another layer.

    Args:
        model (Module): PyTorch model to be edited
        layer_name (string): Name of the layer which will be replaced.
            Dot-notation supported, e.g. "layer10.fc".

    Returns:
        Module: PyTorch model with the new layer set at the specified location.
    """
    setattr_dot_notation(model, layer_name, new_layer)
    return model


def replace_layer_with_identity(model: Module, layer_name: str) -> Module:
    """Replaces any layer of the network with an Identity layer.
    Useful for removing layers of a network to be used as a backbone.

    Args:
        model (Module): PyTorch model to be edited
        layer_name (string): Name of the layer which will be replaced with an
            Identity function. Dot-notation supported, e.g. "layer10.fc".

    Returns:
        Module: PyTorch model with Identity layer at the specified location.
    """
    return replace_layer_with(model, layer_name, Identity())


def remove_n_last_layers_sequentially(model: Module(), num_layers=1) -> Sequential:
    """Remove a number of last layers of a network and return it as an nn.Sequential model.
    Useful when a network is used as a backbone of an SSL model.

    Args:
        model (Module): PyTorch model object.
        num_layers (int, optional): Number of last layers to be removed. Defaults to 1.

    Returns:
        Sequential: PyTorch Sequential model with the last layer removed.
    """
    return Sequential(*list(model.children())[:-num_layers])


def adjust_prefix_and_load_state_dict(
    model: Module, ckpt_path: str, ckpt_to_model_prefix: Dict[str, str] = None, layers_to_ignore: List[str] = None
) -> Module:
    """Load state_dict from a checkpoint into a model using `torch.load(strict=False`).
    `ckpt_to_model_prefix` mapping allows to rename the prefix of the checkpoint's state_dict keys
    so that they match those of the model's state_dict. This is often needed when a model was trained
    as a backbone of another model, so its state_dict keys won't be the same to those of a standalone
    version of that model. Prior to defining the `ckpt_to_model_prefix`, it is advised to manually check
    for mismatch between the names and specify them accordingly.

    Args:
        model (Module): The PyTorch model instance to load the state_dict into.
        ckpt_path (str): Path to the checkpoint.
        ckpt_to_model_prefix (Dict[str, str], optional): A dictionary that maps keys in the checkpoint's
            state_dict to keys in the model's state_dict. If None, no key mapping is performed. Defaults to None.
        layers_to_ignore (List[str], optional): A list of layer names that won't be loaded into the model.
            Specify the names as they are after `ckpt_to_model_prefix` is applied. Defaults to None.
    Returns:
        Module: The model instance with the state_dict loaded.

    Raises:
        ValueError: If there is no overlap between checkpoint's and model's state_dict.
    """

    # Load checkpoint
    ckpt = torch.load(ckpt_path)

    # Check if the checkpoint is a model's state_dict or a LighterSystem checkpoint.
    # A LighterSystem checkpoint contains the model’s entire internal state, we only need its state_dict.
    if "state_dict" in ckpt:
        ckpt = ckpt["state_dict"]
        # Remove the "model." prefix from the checkpoint's state_dict keys. This is characteristic to LighterSystem.
        ckpt = {key.replace("model.", ""): value for key, value in ckpt.items()}

    # Adjust the keys in the checkpoint's state_dict to match the the model's state_dict's keys.
    if ckpt_to_model_prefix is not None:
        for ckpt_prefix, model_prefix in ckpt_to_model_prefix.items():
            # Add a dot at the end of the prefix if necessary.
            ckpt_prefix = ckpt_prefix if ckpt_prefix == "" or ckpt_prefix.endswith(".") else f"{ckpt_prefix}."
            model_prefix = model_prefix if model_prefix == "" or model_prefix.endswith(".") else f"{model_prefix}."
            if ckpt_prefix != "":
                # Replace ckpt_prefix with model_prefix in the checkpoint state_dict
                ckpt = {key.replace(ckpt_prefix, model_prefix): value for key, value in ckpt.items() if ckpt_prefix in key}
            else:
                # Add the model_prefix before the current key name if there's no specific ckpt_prefix
                ckpt = {f"{model_prefix}{key}": value for key, value in ckpt.items() if ckpt_prefix in key}

    # Check if there is no overlap between the checkpoint's and model's state_dict.
    if not set(ckpt.keys()) & set(model.state_dict().keys()):
        raise ValueError(
            "There is no overlap between checkpoint's and model's state_dict. Check their "
            "`state_dict` keys and adjust accordingly using `ckpt_prefix` and `model_prefix`."
        )

    # Remove the layers that are not to be loaded.
    if layers_to_ignore is not None:
        for layer in layers_to_ignore:
            ckpt.pop(layer)

    # Load the adjusted state_dict into the model instance.
    incompatible_keys = model.load_state_dict(ckpt, strict=False)

    # Log the incompatible keys during checkpoint loading.
    if len(incompatible_keys.missing_keys) > 0 or len(incompatible_keys.unexpected_keys) > 0:
        logger.info(f"Encountered incompatible keys during checkpoint loading. If intended, ignore.\n{incompatible_keys}")

    return model



from typing import Any, Callable

import random

from torch.utils.data import DataLoader
from torch.utils.data._utils.collate import collate_str_fn, default_collate_fn_map
from torch.utils.data.dataloader import default_collate

# Collate support for None. Just as a string, None is not collated. Allows elements of the batch to be None.
default_collate_fn_map.update({type(None): collate_str_fn})


def collate_replace_corrupted(batch: Any, dataset: DataLoader, default_collate_fn: Callable = None) -> Any:
    """Collate function that allows to replace corrupted examples in the batch.
    The dataloader should return `None` when that occurs.
    The `None`s in the batch are replaced with other, randomly-selected, examples.

    Args:
        batch (Any): batch from the DataLoader.
        dataset (Dataset): dataset that the DataLoader is passing through. Needs to be fixed
            in place with functools.partial before passing it to DataLoader's 'collate_fn' option
            as 'collate_fn' should only have a single argument - batch. Example:
                ```
                collate_fn = functools.partial(collate_replace_corrupted, dataset=dataset)`
                loader = DataLoader(dataset, ..., collate_fn=collate_fn).
                ```
        default_collate_fn (Callable): the collate function to call once the batch has no corrupted examples.
            If `None`, `torch.utils.data.dataloader.default_collate` is called. Defaults to None.
    Returns:
        Any: batch with new examples instead of corrupted ones.
    """
    # Use `torch.utils.data.dataloader.default_collate` if no other default collate function is specified.
    default_collate_fn = default_collate_fn if default_collate_fn is not None else default_collate
    # Idea from https://stackoverflow.com/a/57882783
    original_batch_len = len(batch)
    # Filter out all the Nones (corrupted examples).
    batch = list(filter(lambda x: x is not None, batch))
    filtered_batch_len = len(batch)
    # Num of corrupted examples.
    num_corrupted = original_batch_len - filtered_batch_len
    if num_corrupted > 0:
        # Replace a corrupted example with another randomly selected example.
        batch.extend([dataset[random.randint(0, len(dataset) - 1)] for _ in range(num_corrupted)])
        # Recursive call to replace the replacements if they are corrupted.
        return collate_replace_corrupted(batch, dataset)
    # Finally, when the whole batch is fine, apply the default collate function.
    return default_collate_fn(batch)
